# Copyright (c) 2004-2009 Moxie Marlinspike
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

import logging
import select, socket, SocketServer, sys
import re,time,string,cgi, traceback

from SecureLinkListener import SecureLinkListener
from ShuffleSocket import ShuffleSocket
from RetrySecureException import RetrySecureException

class DataShuffler:
    '''This class shuffles input across sockets.  As it goes, it watches out for:
    1) https links
    2) https redirects
    3) 404's
    
    '''

    clientSocket       = None
    serverSocket       = None
    secureLinkListener = None
    serverHost         = None
    urlExpression      = None
    notFoundExpression = None

    # Once again, fuck you python.  If I were using a sane programming language,
    # I'd have multiple constructors.  One of them would even, you know, be declared
    # protected.  But to have function overloading you'd have to have, like, static types,
    # so that's just not possible.

    def __init__(self, clientSocket, serverHost, secureLinkListener):
        self.serverHost   = serverHost
        self.serverSocket = ShuffleSocket(host=serverHost)
        DataShuffler.initialize(self, clientSocket, secureLinkListener);

    def initialize(self, clientSocket, secureLinkListener):
        self.clientSocket       = ShuffleSocket(socket=clientSocket)        
        self.secureLinkListener = secureLinkListener
        self.urlExpression      = re.compile(r"(https://[\w\d:#@%/;$()~_?\+-=\\\.&]*)")
        self.notFoundExpression = re.compile(r'^HTTP/1\.[0-1] 404')

    def getRequestLogLevel(self):
        return logging.DEBUG

    def getHeaderLogLevel(self):
        return logging.DEBUG

    def cleanHeaders(self, headers):
        del headers['accept-encoding']
        del headers['keep-alive']
        del headers['connection']
        del headers['if-modified-since']

        headers['Connection'] = 'close'

        return headers        

    def getPostPrefix(self):
        return "POST"

    def sendPotentialData(self, rfile, length):
        data = rfile.read(length)
        rfile.close()

        if data:
            logging.warning(self.getPostPrefix() + " Data (" + self.serverHost + "): " + str(data))
            self.serverSocket.write(data)

    def sendRequest(self, command, path, headers):
        logging.log(self.getRequestLogLevel(),
                    "Sending Request: %s %s %s" % (command, path, "HTTP/1.0"))
        self.serverSocket.write("%s %s %s\r\n" % (command, path, "HTTP/1.0"))

        headers = self.cleanHeaders(headers)

        for key_value in headers.items():
            logging.log(self.getHeaderLogLevel(), "Sending header: %s: %s" % key_value)
            self.serverSocket.write("%s: %s\r\n" % key_value)

        self.serverSocket.write("\r\n")

    def replaceSecureLinks(self, data):
        iterator      = re.finditer(self.urlExpression, data, re.IGNORECASE)

        for match in iterator:
            url = match.group()
            logging.debug("Found secure reference: " + url)

            url = url.replace('https://', 'http://', 1)
            url = url.replace('&amp;', '&')
            self.secureLinkListener.addLink(url);

        return data.replace('https://', 'http://')


    def checkForAbortiveCase(self, data):
        # If we get a 404 on an insecure connection, it could
        # be that we missed something that we should have mapped
        # to SSL.  So we're going to abort, and try connecting out
        # to the same URL via SSL.  If that also returns 404, the
        # user will get a 404 that way.

        if re.search(self.notFoundExpression, data):
            logging.warning("Found 404, returning abort.")
            return True

    def processAndShuffle(self, inStream, outStream):
        data = inStream.read()

        if not data: return False

        if self.checkForAbortiveCase(data):
            raise RetrySecureException, "Got 404 on page."

        data = self.replaceSecureLinks(data)
        outStream.write(data)

        return True

    def shuffleInput(self, input):
        try:
            for sock in input:

                inStream  = None
                outStream = None

                if sock is self.serverSocket._socket: 
                    inStream  = self.serverSocket
                    outStream = self.clientSocket
                else:                                 
                    inStream  = self.clientSocket
                    outStream = self.serverSocket

                return self.processAndShuffle(inStream, outStream)

        except RetrySecureException, arg:
            raise RetrySecureException, arg
        except socket.sslerror:
            return 0
        except:
            logging.debug("Socket exception.  Closing.")
#            traceback.print_exc()
            return 0

    def handleResponse(self):
        readList  = [self.clientSocket._socket, self.serverSocket._socket]
        writeList = []
        count     = 0

        while 1:
            (input, _, exception) = select.select(readList, writeList, readList, 10);

            if exception:
                return False

            if input:     
                if not self.shuffleInput(input):
                    return False
                
            else:
                logging.debug("Select timed out...")
                count+=1
                if count >= 100:
                    return False

    def close(self):
        self.clientSocket.close()
        self.serverSocket.close()
