# Copyright (c) 2004-2009 Moxie Marlinspike
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

import string
import logging
import socket, re
from RetrySecureException import RetrySecureException

class ShuffleSocket:

    '''ShuffleSocket doesn't do much but provide a little custom buffering on
    top of a standard socket.  It doesn't modify anything that it reads or writes,
    and its only responsibility is to return data in pieces that don't break along
    HTML tags, cookies, or CSS expressions.

    '''

    _socket                   = None
    _buffer                   = ""
    _trailingLinkExpression   = None
    _trailingCookieExpression = None
    _trailingCssExpression    = None

    # This, of course, is fucking bullshit.  Since python doesn't
    # have, you know, declared types -- it of course can't support things like
    # function overloading.  Which naturally means that you can't have
    # more than one fucking constructor in a class!  What I'd /like/ to
    # do is have ShuffleSocket implement two constructors: one for connecting
    # and one to take a socket that's already connected.  But I can't do that,
    # so instead I have to do this lame hack where I use default argument values
    # to ridiculously figure out what the caller wants.
    #
    # Fuck you python, it was nice while it lasted but I think I've seen enough.
    # I don't understand how anyone can think this is a language worth using.

    def __init__(self, host=None, port=80, socket=None):
        if host == None: self._socket = socket
        else:            self._socket = self.connectSocket(host, port)

        self._trailingLinkExpression   = re.compile(r'<[\s\w\d:#@%/;$()~_?\+-=\\\.&\"]*$')
        self._trailingCookieExpression = re.compile(r'Set-Cookie:[ \w\d:#@%/;$()~_?\+-=\\\.&]+$')
        self._trailingCssExpression    = re.compile(r'url\([\w\d:#@%/;$~_?\+-=\\\.&]+$')

    def getLogLevel(self):
        return logging.DEBUG

    def connectSocket(self, host, port):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(5.0)
            sock.connect((host, port))
            sock.settimeout(None)

            return sock
        except:
            raise RetrySecureException, "Failed to connect."

    def write(self, data):
        self._socket.send(data)

    def checkForTrailingExpresion(self, expressionList, currentBite):
        currentBite = self._buffer + currentBite
        index       = len(currentBite)
        
        for expression in expressionList:
            match = re.search(expression, currentBite)
            
            if match:
                if match.start() < index:
                    index = match.start()
                    logging.debug("Found match in currentBite: " + currentBite[index:])
        
        if index != len(currentBite): self._buffer = currentBite[index:]
        else:                         self._buffer = ""

        return currentBite[0:index]

    def read(self):
        currentBite = self._raw_read();
        logging.log(self.getLogLevel(), "Read From Server:\n" + currentBite)

        if not currentBite:            
            if not self._buffer == "": toReturn = self._buffer
            else:                      toReturn = currentBite

            self._buffer = ""
            return toReturn
            
        expressionList = [self._trailingLinkExpression, 
                          self._trailingCookieExpression,
                          self._trailingCssExpression]

        return self.checkForTrailingExpresion(expressionList, currentBite)
        
    def _raw_read(self):
        return self._socket.recv(8192)

    def close(self):
        self._socket.close()

    def setblocking(self, val):
        self._socket.setblocking(val)
